# this query selects some standard list processing functions
# and demonstrates the use of some simple higher order processing
# as added by quip.


# standard head and tail functions
DEFINE FUNCTION head (LIST $es) RETURNS ELEMENT
{
	LET $t := <t>{ $es }</t>
	RETURN
		$t/node()[1]
}

DEFINE FUNCTION tail (LIST $es) RETURNS ELEMENT
{
	LET $t := <t>{ $es }</t>
	RETURN
		$t/node()[2 TO last()]
}

# the equivalent functions on the element <li>
DEFINE FUNCTION hd (ELEMENT $t) RETURNS ELEMENT
{
	$t/node()[1]
}
DEFINE FUNCTION tl (ELEMENT $t) RETURNS ELEMENT
{
	<li>{ $t/node()[2 TO last()] }</li>
}

# functions from Clean StdList    
DEFINE FUNCTION isEmpty (ELEMENT $t) RETURNS ELEMENT
{
	empty($t/node())
}

DEFINE FUNCTION cons (ELEMENT $x, ELEMENT $xs) RETURNS ELEMENT
{
	<li>{ $x }{ $xs/node() }</li>
}

DEFINE FUNCTION take (INTEGER $i, LIST $es) RETURNS LIST
{
	LET $t := <t>{ $es }</t>
	RETURN
		$t/node()[1 TO $i]
}

DEFINE FUNCTION drop (INTEGER $i, LIST $es) RETURNS LIST
{
	LET $t := <t>{ $es }</t>
	RETURN
		$t/node()[($i + 1) TO last()]
}

DEFINE FUNCTION reverse (LIST $xs) RETURNS LIST
{
	IF (empty($xs)
	) THEN ()
	ELSE (reverse(tail($xs)), head($xs))
}

DEFINE FUNCTION lastL (LIST $es) RETURNS ELEMENT
{
	LET $t := <t>{ $es }</t>
	RETURN
		$t/node()[last()]
}

DEFINE FUNCTION zipWith (CHARSTRING $t, LIST $xs, LIST $ys) RETURNS LIST
{
	IF (empty($xs) OR empty($ys)
	) THEN ()
	ELSE (<{ $t }>{ head($xs) }{ head($ys) }</{ $t }>, zipWith($t, tail($xs), tail($ys)))
}

DEFINE FUNCTION map (CHARSTRING $f, LIST $xs) RETURNS LIST
{
	IF (empty($xs)
	) THEN ()
	ELSE (apply($f, head($xs)), map($f, tail($xs)))
}

DEFINE FUNCTION plus5 (INTEGER $x) RETURNS INTEGER
{
	$x + 5
}

# merge sort taken from Clean StdOrdList
DEFINE FUNCTION sortItBy (CHARSTRING $f, ELEMENT $xs) RETURNS ELEMENT
{
	hd(msort($f, pair($f, $xs)))
}
DEFINE FUNCTION pair (CHARSTRING $f, ELEMENT $xs) RETURNS ELEMENT
{
	IF (isEmpty($xs)
	) THEN <li>{ $xs }</li>
	ELSE 
	IF (isEmpty(tl($xs))
	) THEN <li>{ $xs }</li>
	ELSE 
	LET $x1 := hd($xs)
	LET $x2 := hd(tl($xs))
	RETURN
		IF (apply($f, $x2, $x1)
		) THEN cons(<li>{ $x2 }{ $x1 }</li>
		, pair($f, tl(tl($xs))))
		ELSE cons(<li>{ $x1 }{ $x2 }</li>
		, pair($f, tl(tl($xs))))
}
DEFINE FUNCTION msort (CHARSTRING $f, ELEMENT $xs) RETURNS ELEMENT
{
	IF (isEmpty($xs)
	) THEN $xs
	ELSE 
	IF (isEmpty(tl($xs))
	) THEN $xs
	ELSE msort($f, merge_stage($f, $xs))
}
DEFINE FUNCTION merge_stage (CHARSTRING $f, ELEMENT $xss) RETURNS ELEMENT
{
	IF (isEmpty($xss)
	) THEN $xss
	ELSE 
	IF (isEmpty(tl($xss))
	) THEN $xss
	ELSE 
	LET $xs1 := hd($xss)
	LET $xstl := tl($xss)
	LET $xs2 := hd($xstl)
	LET $yss := tl($xstl)
	RETURN
		cons(mergeBy($f, $xs1, $xs2), merge_stage($f, $yss))
}
DEFINE FUNCTION mergeBy (CHARSTRING $f, ELEMENT $xs, ELEMENT $ys) RETURNS ELEMENT
{
	IF (isEmpty($xs)
	) THEN $ys
	ELSE 
	IF (isEmpty($ys)
	) THEN $xs
	ELSE 
	IF (apply($f, hd($ys), hd($xs))
	) THEN cons(hd($ys), mergeBy($f, $xs, tl($ys)))
	ELSE cons(hd($xs), mergeBy($f, tl($xs), $ys))
}

# the standard less function 
DEFINE FUNCTION leq (ELEMENT $x, ELEMENT $y) RETURNS BOOLEAN
{
	$x <= $y
}


# example application
LET $y := <li>
             <int>-1</int>
             <int>2</int>
             <int>2</int>
             <int>3</int>
             <int>2</int>
             <int>1</int>
             <int>-15</int>
             <int>17</int>
           </li>
LET $x := sortItBy("leq", $y)

FOR $i IN (map("plus5", $x/*))
RETURN  <int>{$i} </int>

